// -----------------------------------------------------------------
// cardread - reads cards from Documation card reader through
// USB port via SX-based controller.
// -----------------------------------------------------------------

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>

#include "cardread.h"

// -----------------------------------------------------------------
// NOTICES:

#define COPYRIGHT "Copyright 2003 Quarterbyte Systems, Inc."

// Licensed under the GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt.
//
// Copying and distribution of this software is permitted in any medium, provided
// these notices are preserved.
// -----------------------------------------------------------------

#define VERSION "Version 1.3"

// Revision History:
//
// 1.3 20Oct2004	Added option for creating encoding tables with punch-row lists
//
// 1.2 30Jul2004	Added "stop on error" option to enable cardread to resume reading
//					after an error. (Necessitated by odd deck at Computer History Museum
//					with punches in column 81, which stops the Documation reader with a data check).
//					Program now displays text any time card image is displayed, even in Binary mode.
//
// 1.1 07Sep2003	Changed copyright notice above from "Verbatim copying" to "Copying".
//					Added autoverify option.
//					Corrected status message displayed during verify from "Reading..." to "Verifying...".
//					Made "View" button invoke DeckView for binary files, if available.
//					Made registry values for binary and ascii viewer program names so they can be changed
//					Corrected display of card text when an ampersand is present on the card.
//					Added code to permit selection of encoding scheme.
//					Added code to permit debugging of GUI (only) if COM port is set to NUL or DEBUG
//
// -----------------------------------------------------------------
// Note to anyone interested in building one of these: I'd seriously consider modifying the accompanying
// interface circuit to use a USBIO24 module (see e.g. www.dontronics.com) based on the Ubicom SX48, rather than
// the current design's SX28, USBMOD2 and 82C55A. This would be much simpler to build, and would let you view
// all of the Documation's status bits all the time. The USBIO24 uses a 5 MHz clock, not quite fast enough to
// both watch for the Documation's 1 us data strobe pulse AND watch for incoming USB commands. However, if
// you connect the strobe to a port B pin you can have the strobe generate an interrupt. Firmware changes
// would be required since the current design already uses the interrupt for timing.
// -----------------------------------------------------------------

/* The following description of the interface protocol is lifted from the interface
 * firmware source code, file museumread.src (standard Documation interface) or
 * cardread.src (HP differential interface).

; ------------------------------------------------------------------------------		
; Protocol (send through USB using Virtual COM Port driver)
; ------------------------------------------------------------------------------		
; 
; Single-character ASCII commands sent from host computer to the interface
; are followed with a response as indicated in the description of each command.
; The Reset command can be issued at any time. The remaining commands can be
; issued while a pick is pending but not started, or when no pick is pending,
; but not while a card is in the process of being read and its data transmitted,
; that is, not between = and [status]. Commands issued during data transmission
; are ignored.
;
; The commands are listed below.
;
; CMD	DESCRIPTION
; ---   ------------------------------------------------------------------------
;  R	Reset     (can be issued any time)
;		Response: [status]
; 
;		Resets the interface. If the card reader is ready, one card will be read
;		but ignored, due to an unavoidable pulse on the pick-command line.
;
;  S	Status check
;		Response: [status]
;
;		Note: while a pick is pending, some status bits may not be accurate.
;		See the description of [status], below.
; 
;  P	Pick one card
;		Response: [status]
;
;		then, after reading card (which could be quite some time later)
;			  = [card data] [status]
;
;		or, if EOF button is pressed with hopper empty before a card is read
;			  ! [status]
;
;		After reading data, be sure to check the status byte for
;		the error bit, which would indicate a light or dark check -- bad news.
;		Status is sent 1.75 msec after last card column, to give time for
;		hopper check or read check to be set by the reader. Ready, however,
;		will usually not be back on by the time the status byte is sent.
;		For maximum picking rate, if hopper empty and EOF are not set, issue
;		another pick command immedately ater receiving the status byte, if
;		even if ready is not set. You can poll for status afterward to
;		see if error comes up (if you care).
;				
;  X	Cancel pending pick
;		Response: [status]
; 
;  C	Clear EOF lamp. (Best to call only when hopper is empty)
;		Response: [status]
;	
;  (any unrecognized command)
;		Response: ?
;
; NOTE: The program talking to the card reader interface must be sure to read all
; data from one card before a new pick is issued, otherwise the interface may lock
; up. There is insufficient RAM in the interface to buffer more than one card.
; If the FIFO overflows, columns will be missed, and the card data will not
; be received completely. To ensure recovery from an overrun lockup, the calling
; program should set a timeout of no less than 200 msec after receiving the =; if
; 161 bytes aren't received in this time, issue an R reset command and recover.
;
; NOTE: the firmware accepts commands sent in upper or lower case. To see if
; the interface is present, have your program on the PC send a lowercase 's'.
; If there is no response, the interface is not active. If the response is 's',
; you are talking to a modem or something. If the response is a valid staus
; character, you know you are talking to the interface.
;
; ------------------------------------------------------------------------------		
; [status] is a character between 0x40 and 0x5f, that is, @ABC...XYZ[\]^_
; ------------------------------------------------------------------------------		
; 
; bit 0: ready
;     1: error (read check only -- pick check and hopper check are ignored)
;     2: hopper empty
;     3: end of file button pressed
;     4: pick pending
;     5: 0
;     6: 1
;     7: 0
;
; Status character decoder:
;      Pick	   Hpr
; Char Pnd EOF Emp Err Rdy
; ------------------------
;  @	0	0	0	0	0
;  A	0	0	0	0	1
;  B	0	0	0	1	0
;  C	0	0	0	1	1
;  D	0	0	1	0	0
;  E	0	0	1	0	1
;  F	0	0	1	1	0
;  G	0	0	1	1	1
;  H	0	1	0	0	0
;  I	0	1	0	0	1
;  J	0	1	0	1	0
;  K	0	1	0	1	1
;  L	0	1	1	0	0
;  M	0	1	1	0	1
;  N	0	1	1	1	0
;  O	0	1	1	1	1
;  P	1	0	0	0	0
;  Q	1	0	0	0	1
;  R	1	0	0	1	0
;  S	1	0	0	1	1
;  T	1	0	1	0	0
;  U	1	0	1	0	1
;  V	1	0	1	1	0
;  W	1	0	1	1	1
;  X	1	1	0	0	0
;  Y	1	1	0	0	1
;  Z	1	1	0	1	0
;  [	1	1	0	1	1
;  \	1	1	1	0	0
;  ]	1	1	1	0	1
;  ^	1	1	1	1	0
;  _	1	1	1	1	1
;
; NOTE: When a pick is pending, the error and ready bits may not be accurate.
; They will have the same values as when the pick was issued. They will
; be sent correctly after the card had been read, or after the pick is canceled.
;
; Another issue: read checks can occur at the END of the card anywhere between
; columns 81 to 84. The status returned at the end of a read command may not
; reflect an error posted more than 1.75 msec after column 80.
;
; (On an M600, column 80 + 0.87 ms = column 81, HEMPTY may be raised, ERROR may be raised (light check)
;                          1.74             82, [status] byte sent, ERROR may be raised (light check)
;                          2.61             83, ERROR may be raised (light check)
;                          3.48             84, ERROR may be raised (dark check), new pick cycle may start
;
; ------------------------------------------------------------------------------		
; [card data] - 160 bytes of data in the following format
; ------------------------------------------------------------------------------		
;
; byte   0  [ 6  7  8  9  .  .  .  .] column  1 (the four low bits are 0)
; byte   1  [12 11  0  1  2  3  4  5]
; byte   2  [ 6  7  8  9  .  .  .  .] column  2 ( " )
; byte   3  [12 11  0  1  2  3  4  5]
; ...
; byte 158  [ 6  7  8  9  .  .  .  .] column 80 ( " )
; byte 159  [12 11  0  1  2  3  4  5]
;
; The 160 bytes of card data are followed by a status character.
;
; When read as 16-bit integers in x86 word order, the card data appear
; in the format used by the IBM 1130 simulator:
;  MSB									      LSB
; +----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
; | 12 | 11 |  0 |  1 |  2 |  3 |  4 |  5 |  6 |  7 |  8 |  9 |    |    |    |    |
; +----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
;
; The low four bits are zero.
; ------------------------------------------------------------------------------


Cardread has built-in tables for the IBM 029, IBM 026 Fortran and IBM 026 Commercial
encoding schemes. New encodings can be defined (or the default encodings replaced)
by creating a file named "cardcode.tbl" in the same directory as cardread.exe. The
structure of cardcode.tbl is as follows:

* blank lines and lines starting with /, *, # or ; are ignored as comments

* A new encoding scheme is introduced by a line containing the name of the
  scheme enclosed in square brackets, e.g. [029] or [My Special]. It's best if the
  name is 12 characters or less, so it will display properly in the dropdown box.

* After the name there may be up to 150 encoding entries. Each consists of a four
  character hexadecimal number listing the hollerith code LEFT JUSTIFED in 16 bits
  (e.g. in cardread's binary format), followed by whitespace, followed by either a single
  non-white ascii character OR by a two-digit hexadecimal number providing the ASCII value.

* The character or character code may optionally be followed by whitespace and comment text.

* Ordering of hollerith/ascii code pairs within a given code table is not important, although
  it can't hurt to place the most common codes near the start of the table, as it is searched
  sequentially.

For example, the 029 encoding might start like this:

	[029]
	0000	20	  	space
	8000 	&
	4000	-
	2000	0
	1000	1
	0800	2
	0400	3

and so on. The encoding table names appear in the dropdown box in
CardRead's "Data" section.

The correspondence between the hexadecimal hollerith values and
the card punch rows are listed below:

  Value   Punch Row
  ------  ---------
  8000       12
  4000	     11
  2000	      0
  1000	      1
  0800	      2
  0400	      3
  0200	      4
  0100	      5   
  0080	      6
  0040	      7
  0020	      8
  0010	      9

*/

/* program note:
 *
 * Flag "am_reading" tells whether the GUI thread or the file reading thread is in
 * charge of the USB interface. Only one or the other is allowed to talk to the interface,
 * otherwise IO might get interleaved.
 *
 * Main GUI thread normally talks to the interface (e.g. periodic status checks
 * triggered by Timer messages). When user clicks Start, GUI thread sets "am_reading,"
 * disables most buttons by calling enable_controls(), and signals an event to release
 * the read thread. The read thread then can talk to the interface as desired.
 * When it's done, it clears am_reading, calls enable_controls() and waits for the event
 * again.
 *
 * The GUI thread can interrupt the reading by setting "stop_pending," which the
 * read thread will see (eventually, within 400 msec or less) and quit.
 */

#define STS_READY		0x01					// status byte bit decoder
#define STS_ERROR		0x02
#define STS_EMPTY		0x04
#define STS_EOF			0x08
#define STS_PENDING		0x10

#define DEFAULT_PORT			"COM3"
#define DEFAULT_SEPSPEC			"1:6-7-8-9"
#define DEFAULT_ASCII_VIEWER	"notepad.exe"
#define DEFAULT_BINARY_VIEWER	"deckview.exe"
#define DEFAULT_ENCODING_FILE	"cardcode.tbl"
#define DEFAULT_ENCODING_NAME	"029"

#define BAD_HOLLERITH	0xFFFF					// invalid Hollerith code

#define REGISTRY_KEY "Software\\Quarterbyte\\CardRead"

#define ID_TIMER		1

// -----------------------------------------------------------------
// Global Variables
// -----------------------------------------------------------------

char port[MAX_PATH] = "";						// default port
BOOL comm_ok  =  FALSE;							// if TRUE, port is opened and reader is talking
BOOL configuring_port = FALSE;					// if TRUE, we are using the Select Port dialog
BOOL verbose  = FALSE;							// if TRUE, we are debugging
BOOL display  = FALSE;							// if TRUE, we are displaying binary card image
BOOL have_data = FALSE;							// if TRUE, we have a card in the cardbuf and can display it
BOOL onecard  = FALSE;							// if TRUE, we are to read and display one card exactly
BOOL stoponerror = TRUE;						// if TRUE, we close the output file if there is a card data error
HANDLE hPort  = INVALID_HANDLE_VALUE;
BOOL to_ascii = TRUE;							// if TRUE, we are to convert to ASCII
BOOL verify   = FALSE;							// if TRUE, this read is a verify run
FILE *fout    = NULL;							// current output file
char status   = '@';							// last-seen status value from the reader interface
int nrec	  = 0;								// number of cards read in current file
BOOL numup    = FALSE;							// if TRUE, the remark window contains card count string
BOOL autosep  = FALSE;							// if TRUE, autosep mode is enabled
BOOL autoincr = FALSE;							// if TRUE, autoincrement mode is enabled
BOOL autoverify = FALSE;						// if TRUE, autoverify mode is enabled
BOOL did_exist  = FALSE;						// if TRUE, file we are going to read existed before we started
int  startup    = 3;							// counts down initial copyright display before enabling controls
char curname[MAX_PATH] = "CARD0000.DAT";	 	// default filename
char ascii_viewer[MAX_PATH];					// program to use for viewing ascii files
char binary_viewer[MAX_PATH];					// program to use for viewing binary files
char encoding_file[MAX_PATH];					// optional file listing alternate encodings
char ascii_viewer_path[MAX_PATH];				// full path to specified programs
char binary_viewer_path[MAX_PATH];				// (if found)
char encoding_file_path[MAX_PATH];				// full path to encoding file (if found)
USHORT cardbuf[80];								// last-read card (80 16-bit)
USHORT vbuf[80];								// last-read binary data from file during verification
char cardtext[90];								// converted text copy of cardbuf (leave 10 extra spaces)
BOOL is_autosep;								// if TRUE, the current card is an autosep card
char sepspec[160];								// specification string for autosep card
char encoding_name[50];							// current encoding name
USHORT sepbuf[80];								// image of autosep card
BOOL am_reading = FALSE;						// if TRUE, read is in progress -- indicates that read_thread is in
												// charge of the USB interface.
BOOL do_autoincr   = FALSE;						// if TRUE, restart next file with next filename
BOOL do_autoverify = FALSE;						// if TRUE, verify next file automatically
BOOL stop_pending  = FALSE;						// if TRUE, read is being canceled
BOOL debug_gui     = FALSE;						// if TRUE, we don't try to talk to interface -- used for debugging GUI code
HWND hwCard;									// window handle of rectangle control that we use for card display
HWND hDlgMain;									// window handle of the main dialog
HINSTANCE ghInstance;							// instance handle of program
HANDLE hThread  = NULL;							// handle to read thread
HANDLE hEvent   = NULL;							// handle to synchronization event object
HBRUSH bManilla = NULL;							// handle to a brush the color of a manilla punched card
HBRUSH bRed     = NULL;							// handle to a red brush
HBRUSH bBlack   = NULL;							// handle to a red pen
HPEN   pRed		= NULL;							// red pen
WNDPROC oldRectProc;							// original window procedure for the IDC_CARD display control

// -----------------------------------------------------------------
// Prototypes
// -----------------------------------------------------------------

BOOL CALLBACK	CardReadDlgProc				(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL			CardRead_OnInitDialog		(HWND hDlg, HWND hwndFocus, LPARAM lParam);
void 			CardRead_OnCommand			(HWND hDlg, int id, HWND hWndCtl, UINT codeNotify);
void			CardRead_OnSysCommand 		(HWND hDlg, int id, int x, int y);
void			CardRead_OnClose			(HWND hDlg);
void			CardRead_OnTimer			(HWND hDlg, UINT id);

BOOL CALLBACK	ConfigurePortDlgProc		(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL			ConfigurePort_OnInitDialog	(HWND hDlg, HWND hwndFocus, LPARAM lParam);
void			ConfigurePort_OnCommand		(HWND hDlg, int id, HWND hWndCtl, UINT codeNotify);

void			open_comm_port	(void);
void			remark			(char *fmt, ...);
void			CenterMe		(HWND hWnd);
void			load_reg_bool	(HKEY hKey, LPSTR valname, BOOL *val, BOOL default_val);
void			load_reg_str	(HKEY hKey, LPSTR valname, LPSTR val, int maxval, LPSTR default_val);
void			save_reg_bool	(HKEY hKey, LPSTR valname, BOOL val);
void			save_reg_str	(HKEY hKey, LPSTR valname, LPSTR val);
void			get_encodings	(void);
void			get_defaults	(void);
void			save_defaults	(void);
BOOL			locate_program	(char *name, char *fullpath, int s_fullpath);
void			EnableControl	(HWND hDlg, int id, BOOL enab);
void			update_status	(void);
char *			alltrim			(char *str);
char *			astring			(char *str);
char *			strim			(char *str);
char *			skipbl			(char *str);
void			enable_controls	(void);
void			set_display_mode(BOOL newmode, BOOL do_paint);
void			paint_card		(void);
void			show_card		(BOOL show_data, BOOL show_text);
void			display_cardtext(char *txt);
void			convert_card	(void);
BOOL			issue_cmd		(char cmd);
BOOL			valid_status	(char status);
BOOL			read_data		(void);
void			save_data		(void);
void			browse_file		(void);
void			read_file		(BOOL verify_mode);
void			increment_name	(void);
void			next_file		(BOOL verify_mode);
void			view_file		(void);
BOOL			parse_sepspec	(void);
void			bail			(char *msg);
void			stop_reading	(BOOL from_EOF);
DWORD CALLBACK	read_thread		(LPDWORD param);
void			read_one_file	(void);
LRESULT CALLBACK MyRectProc		(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

// -----------------------------------------------------------------
// Default encoding tables & handling routines
// -----------------------------------------------------------------

#include "cr_encoding.h"

ENCODING *cur_encoding = NULL;			// current encoding table

// -----------------------------------------------------------------
// WinMain - main program
// -----------------------------------------------------------------

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	ghInstance = hInstance;
	DialogBox(hInstance, "CARDREAD", NULL, CardReadDlgProc);
	return 0;
}

// -----------------------------------------------------------------
// CardReadDlgProc - dialog procedure for the main CardRead window
// -----------------------------------------------------------------

BOOL CALLBACK CardReadDlgProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			return HANDLE_WM_INITDIALOG(hDlg, wParam, lParam, CardRead_OnInitDialog);

		case WM_COMMAND:
			HANDLE_WM_COMMAND(hDlg, wParam, lParam, CardRead_OnCommand);
			break;

		case WM_SYSCOMMAND:
			HANDLE_WM_SYSCOMMAND(hDlg, wParam, lParam, CardRead_OnSysCommand);
			return FALSE;						// return FALSE so DefDlgProc performs standard processing

		case WM_CLOSE:
			HANDLE_WM_CLOSE(hDlg, wParam, lParam, CardRead_OnClose);
			break;

		case WM_TIMER:
			HANDLE_WM_TIMER(hDlg, wParam, lParam, CardRead_OnTimer);
			break;

		default:
			return FALSE;
	}

	return TRUE;
}

// -----------------------------------------------------------------
// CardRead_OnInitDialog - initialize program
// -----------------------------------------------------------------

BOOL CardRead_OnInitDialog (HWND hDlg, HWND hwndFocus, LPARAM lParam)
{
	HMENU hMenu;
	HICON hIcon;

	if ((hIcon = LoadIcon(ghInstance, "cardread")) != NULL)
		SetClassLong(hDlg, GCL_HICON, (LONG) hIcon);		// try to set the Dialog class icon to our program's icon

	CenterMe(hDlg);											// center dialog on screen

	bManilla = CreateSolidBrush(RGB(234,241,163)); 			// create brush for painting card image
	bRed     = CreateSolidBrush(RGB(255, 0, 0));			// create brush for displaying verify errors
	bBlack   = GetStockObject(BLACK_BRUSH);					// use stock black brush
	pRed     = CreatePen(PS_SOLID, 1, RGB(255, 0, 0));		// create a red pen

	hDlgMain = hDlg;										// save dialog handle
	comm_ok = FALSE;										// communication is not active yet

	get_defaults();											// get previous settings from registry

	get_encodings();										// set up encoding tables

	parse_sepspec();										// initialize separator ONLY after setting up encoding table(s)

	update_status();										// disable status display

	hwCard = GetDlgItem(hDlg, IDC_CARD);					// get handle to rectangle that will contain card picture
	oldRectProc = (WNDPROC) GetWindowLong(hwCard, GWL_WNDPROC);	// superclass this control so we can intercept WM_PAINT messages
	SetWindowLong(hwCard, GWL_WNDPROC, (LONG) (WNDPROC) MyRectProc);
	
															// set controls to match initial settings
	display_cardtext("");
	CheckRadioButton(hDlg, IDC_BINARY,     IDC_ASCII, to_ascii ? IDC_ASCII : IDC_BINARY);
	CheckDlgButton(hDlg,   IDC_AUTOSEP,    autosep    ? 1 : 0);
	CheckDlgButton(hDlg,   IDC_AUTOVERIFY, autoverify ? 1 : 0);
	SetDlgItemText(hDlg,   IDC_SEPSPEC,    sepspec);
	CheckRadioButton(hDlg, IDC_STOP, IDC_AUTOINCR, autoincr ? IDC_AUTOINCR : IDC_STOP);
	SetDlgItemText(hDlg,   IDC_FILENAME,   curname);
	CheckDlgButton(hDlg,   IDC_DISPLAY,    display ? 1 : 0);// do this last so it shows correctly if autosep display turned it on

	enable_controls();										// not really needed as change messages occured as a result of set calls

	hMenu = GetSystemMenu(hDlg, FALSE);						// add the Configure item to the system menu
	AppendMenu(hMenu, MF_SEPARATOR, 0, 0);
	AppendMenu(hMenu, MF_ENABLED, IDM_CONFIGURE, "&Configure");

	set_display_mode(display, FALSE);						// enable/disable binary card display

	remark(VERSION " - " COPYRIGHT);
	SetTimer(hDlg, ID_TIMER, 500, NULL);					// start timer messages every 500 msec
	return TRUE;
}

// -----------------------------------------------------------------
// CardRead_OnCommand - handle command messages (GUI clicks)
//
// (Note: it's not usual Windows programming style, but in this
//  program, ALL dialog values are extracted from the dialog to global
//  variables IMMEDIATELY upon their being changed by the user.
//  This made the port from the command line version easier.)
// -----------------------------------------------------------------

void CardRead_OnCommand (HWND hDlg, int id, HWND hWndCtl, UINT codeNotify)
{
	int ind;

	switch (id) {
		case IDC_ASCII:
		case IDC_BINARY:
			to_ascii = IsDlgButtonChecked(hDlg, IDC_ASCII);

			if (display && have_data) {							// if card display is up, ensure correct text display (on or off)
//				if (to_ascii) {
// Ver 1.2: always display ascii text
					convert_card();								// in ascii mode, be sure we have current text, then display it
					display_cardtext(cardtext);
//				}
//				else											// in binary mode, don't display text
//					display_cardtext("");
			}
			enable_controls();									// possibly update view button
			break;

		case IDC_ENCODING:
			if (codeNotify == CBN_SELCHANGE) {					// change in encoding selection
				if ((ind = SendDlgItemMessage(hDlg, IDC_ENCODING, CB_GETCURSEL, 0, 0)) == CB_ERR)
					ind = 0;
				cur_encoding = &encoding[ind];					// set pointer and get name
				SendDlgItemMessage(hDlg, IDC_ENCODING, CB_GETLBTEXT, ind, (LPARAM) encoding_name);
// Ver 1.2: update text display
				if (have_data) {
					convert_card();
					display_cardtext(cardtext);
				}
			}			
			break;

		case IDC_DISPLAY:
			set_display_mode(IsDlgButtonChecked(hDlg, IDC_DISPLAY), TRUE);
			break;

		case IDC_AUTOSEP:
			autosep = IsDlgButtonChecked(hDlg, IDC_AUTOSEP);
			if (autosep) {										// autosep negates autoverify
				CheckDlgButton(hDlg, IDC_AUTOVERIFY, FALSE);
				autoverify = FALSE;
			}
			enable_controls();									// enable/disable other autosep input fields
			if (autosep)
				parse_sepspec();
			break;

		case IDC_AUTOVERIFY:
			autoverify = IsDlgButtonChecked(hDlg, IDC_AUTOVERIFY);
			if (autoverify) {									// autoverify negates autosep
				CheckDlgButton(hDlg, IDC_AUTOSEP, FALSE);
				autosep = FALSE;
				enable_controls();
			}
			break;

		case IDC_AUTOINCR:
		case IDC_STOP:
			autoincr = IsDlgButtonChecked(hDlg, IDC_AUTOINCR);	// enable/disable autoincrement mode
			break;

		case IDC_SEPSPEC:
			if (codeNotify == EN_CHANGE || codeNotify == EN_SETFOCUS) {
				GetDlgItemText(hDlg, IDC_SEPSPEC, sepspec, sizeof(sepspec));
				parse_sepspec();							// display it as we go
			}
			else if (codeNotify == EN_KILLFOCUS) {
				if (! parse_sepspec()) {
					MessageBox(hDlg, "Separator specification is invalid", NULL, MB_OK|MB_ICONSTOP);
					SetFocus(GetDlgItem(hDlg, IDC_SEPSPEC));
				}
			}
			break;

		case IDC_FILENAME:
			if (codeNotify == EN_CHANGE) {					// disable OK if port name is blank
				GetDlgItemText(hDlg, IDC_FILENAME, curname, sizeof(curname));
				SetDlgItemText(hDlg, IDC_COUNT,  "");		// remove previous count & remark
				remark("");
				curname[sizeof(curname)-1] = '\0';
				alltrim(curname);
				enable_controls();
			}
			break;

		case IDC_BROWSE:
			browse_file();
			break;

		case IDC_VIEW:
			view_file();
			break;

		case IDC_GO:										// go button works like EOF when reading
			if (am_reading)
				stop_reading(FALSE);
			else
				read_file(FALSE);
			break;

		case IDC_VERIFY:
			read_file(TRUE);
			break;

		case IDC_NEXT:
			next_file(FALSE);
			break;

		case IDCANCEL:										// Esc key acts like Stop button
			if (am_reading)
				stop_reading(FALSE);
			break;
	}
}

// -----------------------------------------------------------------
// CardRead_OnSysCommand - handle system menu command messages
// -----------------------------------------------------------------

void CardRead_OnSysCommand (HWND hDlg, int id, int x, int y)
{
	switch (id) {
		case IDM_CONFIGURE:									// run the ConfigurePort dialog
			if (startup)									// ignore this during 2 seconds of startup phase
				break;

			configuring_port = TRUE;						// halt the period attempts to open the port

			if (DialogBox(ghInstance, "CONFIGUREPORT", hDlg, ConfigurePortDlgProc))
				open_comm_port();							// if changed, reopen port to interface

			configuring_port = FALSE;
			break;
	}
}

// -----------------------------------------------------------------
// CardRead_OnClose - shut down the program
// -----------------------------------------------------------------

void CardRead_OnClose (HWND hDlg)
{
	if (am_reading)											// ignore close when deck read is in progress
		MessageBeep(0);
	else {
		save_defaults();
		EndDialog(hDlg, 0);
	}
}

// -----------------------------------------------------------------
// CardRead_OnTimer - handle periodic timer messages
// -----------------------------------------------------------------

void CardRead_OnTimer (HWND hDlg, UINT id)
{
	static int delay = 4;

	if (startup) {								// during startup phase
		if (--startup == 0)						// tick down the copyright display clock, then
			open_comm_port();					// open the default COM port

		return;
	}

	if (comm_ok && ! am_reading) {
		issue_cmd('S');							// issue "get status" commands twice per second (async with reading)
	}
	else if (! comm_ok) {						// if interface wasn't responding, keep trying every few seconds so it will
		if (configuring_port)					// don't try to do open if we're in the Select Port dialog, though
			delay = 4;
		else if (--delay <= 0) {				// come online if it was unplugged and gets plugged in.
			open_comm_port();
			delay = 4;
		}
	}
}

// -----------------------------------------------------------------
// bail - report fatal error and close the application
// -----------------------------------------------------------------

void bail (char *msg)
{
	MessageBox(hDlgMain, msg, NULL, MB_ICONASTERISK);// display message

	am_reading = FALSE;								// turn off am_reading so application can close
	SendMessage(hDlgMain, WM_CLOSE, 0, 0);
}

// -----------------------------------------------------------------
// remark - format and display a message in the Remark field. Use like printf.
// -----------------------------------------------------------------

void remark (char *fmt, ...)
{
	va_list args;
	char msg[256];

	va_start(args, fmt);							// get pointer to argument list
	vsprintf(msg, fmt, args);						// construct formatted message
	va_end(args);

	SetDlgItemText(hDlgMain, IDC_REMARK, msg);		// display message
}

// -----------------------------------------------------------------
// load_reg_bool - load a boolean value from the registry
// -----------------------------------------------------------------

void load_reg_bool (HKEY hKey, LPSTR valname, BOOL *val, BOOL default_val)
{
	DWORD rtype, dval, len;

	if (hKey != INVALID_HANDLE_VALUE) {
		len = sizeof(dval);
		if (RegQueryValueEx(hKey, valname, NULL, &rtype, (LPBYTE) &dval, &len) == ERROR_SUCCESS) {
			if (rtype == REG_DWORD) {
				*val = dval;
				return;
			}
		}
	}

	*val = default_val;
}

// -----------------------------------------------------------------
// load_reg_str - load a string value from the registry
// -----------------------------------------------------------------

void load_reg_str (HKEY hKey, LPSTR valname, LPSTR val, int maxval, LPSTR default_val)
{
	DWORD rtype, len;

	if (hKey != INVALID_HANDLE_VALUE) {
		len = maxval;
		if (RegQueryValueEx(hKey, valname, NULL, &rtype, val, &len) == ERROR_SUCCESS) {
			if (rtype == REG_SZ)
				return;
		}
	}

	strncpy(val, default_val, maxval);
}

// -----------------------------------------------------------------
// save_reg_bool - save a boolean value to the registry
// -----------------------------------------------------------------

void save_reg_bool (HKEY hKey, LPSTR valname, BOOL val)
{
	DWORD dval;

	if (hKey != INVALID_HANDLE_VALUE) {
		dval = val;
		RegSetValueEx(hKey, valname, 0, REG_DWORD, (CONST BYTE *) &dval, sizeof(DWORD));
	}
}

// -----------------------------------------------------------------
// save_reg_str - save a string value to the registry
// -----------------------------------------------------------------

void save_reg_str (HKEY hKey, LPSTR valname, LPSTR val)
{
	if (hKey != INVALID_HANDLE_VALUE) {
		RegSetValueEx(hKey, valname, 0, REG_SZ, val, strlen(val)+1);
	}
}

// -----------------------------------------------------------------
// locate_program - find a program specified by name and get its full path
// -----------------------------------------------------------------

BOOL locate_program (char *name, char *fullpath, int s_fullpath)
{
	char *basename;

	if (strchr(name, ':') != NULL || strchr(name, '\\') != NULL) {	// user specified an explicit path
		strcpy(fullpath, name);										// use that
	}
	else if (SearchPath(NULL, name, NULL, s_fullpath, fullpath, &basename) == 0) {
		*fullpath = '\0';											// ... not found. zap string and return
		return FALSE;
	}

	// find the program

	if (GetFileAttributes(fullpath) != 0xFFFFFFFF) 					// but be sure file exists
		return TRUE;												// ... it does

	*fullpath = '\0';												// ... it doesn't; zap the path string
	return FALSE;
}

// -----------------------------------------------------------------
// get_encodings - see if there is an encoding table & read it if so
// -----------------------------------------------------------------

void get_encodings (void)
{
	int i, ind;
	ENCODING *e = NULL;
	BOOL selected;

	read_encoding_tables(encoding_file_path, hDlgMain);

	if (nencodings <= 0) {											// if no encodings survived, turn off ascii mode
		to_ascii = FALSE;
		return;
	}

	selected = FALSE;

	for (i = 0; i < nencodings; i++) {								// populate the dropdown list
		if (strlen(encoding[i].name) >= sizeof(encoding_name))		// clip names so any of them can be held in encoding_name
			encoding[i].name[sizeof(encoding_name)-1] = '\0';

		ind = SendDlgItemMessage(hDlgMain, IDC_ENCODING, CB_INSERTSTRING, (WPARAM) -1, (LPARAM) encoding[i].name);
		if (strcmpi(encoding[i].name, encoding_name) == 0) {		// set selection if it matches the default name
			SendDlgItemMessage(hDlgMain, IDC_ENCODING, CB_SETCURSEL, ind, 0);
			cur_encoding = encoding+i;
			selected = TRUE;
		}
	}

	if (! selected) {												// if default name didn't match, select first entry
		SendDlgItemMessage(hDlgMain, IDC_ENCODING, CB_SETCURSEL, 0, 0);
		cur_encoding = &encoding[0];
	}
}

// -----------------------------------------------------------------
// get_defaults - load default values from the registry
// -----------------------------------------------------------------

void get_defaults (void)
{
	HKEY hkReg;

	if (RegOpenKey(HKEY_CURRENT_USER, REGISTRY_KEY, &hkReg) != ERROR_SUCCESS)
		hkReg = INVALID_HANDLE_VALUE;

	load_reg_str( hkReg, "port",     	  port,      	 sizeof(port),    		DEFAULT_PORT);
	load_reg_str( hkReg, "ascii_viewer",  ascii_viewer,  sizeof(ascii_viewer),  DEFAULT_ASCII_VIEWER);
	load_reg_str( hkReg, "binary_viewer", binary_viewer, sizeof(binary_viewer), DEFAULT_BINARY_VIEWER);
	load_reg_str( hkReg, "encoding_file", encoding_file, sizeof(encoding_file), DEFAULT_ENCODING_FILE);
	load_reg_bool(hkReg, "to_ascii", 	  &to_ascii,                  			TRUE);
	load_reg_bool(hkReg, "autosep",  	  &autosep,                  			FALSE);
	load_reg_bool(hkReg, "autoincr", 	  &autoincr,							FALSE);
	load_reg_bool(hkReg, "display",  	  &display,                   			FALSE);
	load_reg_str( hkReg, "sepspec",  	  sepspec,   	 sizeof(sepspec), 		DEFAULT_SEPSPEC);
	load_reg_str( hkReg, "encoding",	  encoding_name, sizeof(encoding_name),	DEFAULT_ENCODING_NAME);
	load_reg_str( hkReg, "filename", 	  curname,   	 sizeof(curname), 		"");
	load_reg_bool(hkReg, "stoponerror",	  &stoponerror,							TRUE);

	if (hkReg != INVALID_HANDLE_VALUE)
		RegCloseKey(hkReg);

	locate_program(ascii_viewer,  ascii_viewer_path,  sizeof(ascii_viewer_path));
	locate_program(binary_viewer, binary_viewer_path, sizeof(binary_viewer_path));
	locate_program(encoding_file, encoding_file_path, sizeof(encoding_file_path));
}

// -----------------------------------------------------------------
// save_defaults - store all GUI settings to registry for next time
// -----------------------------------------------------------------

void save_defaults (void)
{
	HKEY hkReg;
	DWORD dispos;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, REGISTRY_KEY, 0, "",
			REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkReg, &dispos) != ERROR_SUCCESS)
		hkReg = INVALID_HANDLE_VALUE;

	save_reg_str( hkReg, "port",     	  port);
	save_reg_str( hkReg, "ascii_viewer",  ascii_viewer);	// write even though never altered so setting & default value are listed in registry
	save_reg_str( hkReg, "binary_viewer", binary_viewer);
	save_reg_bool(hkReg, "to_ascii",	  to_ascii);
	save_reg_bool(hkReg, "autosep",		  autosep);
	save_reg_bool(hkReg, "autoincr",	  autoincr);
	save_reg_bool(hkReg, "autoverify",	  autoverify);
	save_reg_bool(hkReg, "display",		  display);
	save_reg_str( hkReg, "sepspec",		  sepspec);
	save_reg_str( hkReg, "filename",	  curname);
	save_reg_str( hkReg, "encoding",	  encoding_name);
	save_reg_bool(hkReg, "stoponerror",	  stoponerror);

	if (hkReg != INVALID_HANDLE_VALUE)
		RegCloseKey(hkReg);
}

// -----------------------------------------------------------------
// EnableControl - enable/disable the specified dialog control
// -----------------------------------------------------------------

void EnableControl (HWND hDlg, int id, BOOL enab)
{
	EnableWindow(GetDlgItem(hDlg, id), enab && ! startup);
}

// -----------------------------------------------------------------
// enable_controls - enable or disable controls depending on current context
// (Note: EnableControl disables controls during startup phase)
// -----------------------------------------------------------------

void enable_controls ()
{
	BOOL has_filename, file_exists = FALSE;
	DWORD attr;

	has_filename = *curname;

	if (has_filename) {
		if ((attr = GetFileAttributes(curname)) != 0xFFFFFFFF) {
			file_exists = (attr & FILE_ATTRIBUTE_DIRECTORY) == 0;
		}
	}

	EnableControl(hDlgMain, IDC_GO	,	  ! stop_pending);
	EnableControl(hDlgMain, IDC_ASCII,	  nencodings > 0 && ! am_reading);
	EnableControl(hDlgMain, IDC_BINARY,	  ! am_reading);
	EnableControl(hDlgMain, IDC_ENCODING, nencodings > 0 && to_ascii && ! am_reading);

	EnableControl(hDlgMain, IDC_FILENAME, ! am_reading);	
	EnableControl(hDlgMain, IDC_VIEW,     file_exists  && ! am_reading && (to_ascii ? (*ascii_viewer_path) : (*binary_viewer_path)));
	EnableControl(hDlgMain, IDC_BROWSE,	  ! am_reading);

	EnableControl(hDlgMain, IDC_AUTOSEP,  ! am_reading);
	EnableControl(hDlgMain, IDC_SEPSPEC,  autosep && ! am_reading);
	EnableControl(hDlgMain, IDC_AUTOINCR, TRUE);
	EnableControl(hDlgMain, IDC_DISPLAY,  TRUE);

	EnableControl(hDlgMain, IDC_GO,       has_filename && comm_ok);
	EnableControl(hDlgMain, IDC_VERIFY,   file_exists  && comm_ok && ! am_reading);
	EnableControl(hDlgMain, IDC_NEXT,     has_filename && comm_ok && ((status & STS_EMPTY) || ! am_reading));
}

// -----------------------------------------------------------------
// update_status - update display of status indicators
// -----------------------------------------------------------------

void update_status (void)
{
	int sts = (comm_ok ? status : 0);						// if comm port not working, assume all bits off

	EnableControl(hDlgMain, IDC_HCHK,    sts & STS_EMPTY);	// gray out status indicators for which bit is 0
	EnableControl(hDlgMain, IDC_ERROR,   sts & STS_ERROR);
	EnableControl(hDlgMain, IDC_PENDING, sts & STS_PENDING);
	EnableControl(hDlgMain, IDC_EOF,	 sts & STS_EOF);
}

// -----------------------------------------------------------------
// open_comm_port - attempt to contact the card reader
// -----------------------------------------------------------------

void open_comm_port (void)
{
	DCB dcb;
	COMMTIMEOUTS cto;
	DWORD nerr;
	LPSTR lpMessageBuffer;
	char title[200];

	if (hPort != INVALID_HANDLE_VALUE) {					// close current port
		CloseHandle(hPort);
		hPort = INVALID_HANDLE_VALUE;
	}

	debug_gui = strcmpi(port, "NUL") == 0 || strcmpi(port, "NUL:") == 0 || strnicmp(port, "DEBUG", 5) == 0;
	if (debug_gui) {										// if debugging, just enable controls and return
		issue_cmd('r');
		comm_ok = TRUE;
		remark("");
		enable_controls();
		SetWindowText(hDlgMain, "CardRead - DEBUGGING");
		return;
	}

	comm_ok = FALSE;
	enable_controls();

//	if (strchr(port, ':') == NULL)	  						// colon is not needed?
//		strcat(port, ":");

	sprintf(title, "CardRead - %s", port);					// put port into window title
	SetWindowText(hDlgMain, title);

	remark("Attempting to contact card reader on port %s", port);

	hPort = CreateFile(port, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);

	if (hPort == INVALID_HANDLE_VALUE) {					// explain why it failed
		FormatMessage(
		  FORMAT_MESSAGE_ALLOCATE_BUFFER |
		  FORMAT_MESSAGE_FROM_SYSTEM,
		  NULL,
		  GetLastError(),
		  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 		// The user default language
		  (LPTSTR) &lpMessageBuffer,
		  0,
		  NULL );

		remark("ERROR - %s: %s", port, lpMessageBuffer);
		LocalFree(lpMessageBuffer);
		return;
	}

	memset(&dcb, 0, sizeof(dcb));

	dcb.DCBlength		= sizeof(DCB);
	dcb.BaudRate		= CBR_115200;
    dcb.fBinary  		= 1;
    dcb.fParity	  		= 0;
    dcb.fOutxCtsFlow    = 0;
    dcb.fOutxDsrFlow    = 0;
    dcb.fDtrControl		= DTR_CONTROL_ENABLE;
    dcb.fDsrSensitivity = FALSE;
    dcb.fTXContinueOnXoff = 0;
    dcb.fOutX			= 0;
    dcb.fInX			= 0;
    dcb.fErrorChar		= 0;
    dcb.fNull			= 0;
    dcb.fRtsControl		= RTS_CONTROL_ENABLE;                
    dcb.fAbortOnError   = 0;
    dcb.XonLim			= 0;
    dcb.XoffLim			= 0;
    dcb.ByteSize		= 8;
    dcb.Parity			= NOPARITY;
    dcb.StopBits		= ONESTOPBIT;
    dcb.XonChar			= 0;
    dcb.XoffChar		= 0;
    dcb.ErrorChar		= 0;
    dcb.EofChar			= 0;
    dcb.EvtChar			= 0;

	if (! SetCommState(hPort, &dcb)) {
		CloseHandle(hPort);
		remark("ERROR - %s: SetCommState failed", port);
		return;
	}

    cto.ReadIntervalTimeout         = 100;					// stop if 100 msec elapses between two received bytes
    cto.ReadTotalTimeoutMultiplier  = 0;					// no length sensitivity
    cto.ReadTotalTimeoutConstant    = 400;					// allow 400 msec for a read (reset command can take a while)

    cto.WriteTotalTimeoutMultiplier = 0;
    cto.WriteTotalTimeoutConstant   = 200;					// allow 200 msec for a write

	if (! SetCommTimeouts(hPort, &cto)) {
		CloseHandle(hPort);
		remark("ERROR - %s: SetCommTimeouts failed", port);
		return;
	}

	PurgeComm(hPort, PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
	ClearCommError(hPort, &nerr, NULL);

	if (issue_cmd('r')) {									// reset the interface & expect valid response
		comm_ok = TRUE;										// (use lowercase r so, if connected to a modem or something,
		remark("");											// we get an invalid response back)
		enable_controls();									// enable buttons
	}
}

// -----------------------------------------------------------------
// parse_sepspec - parse the separator specification into a binary card image
// -----------------------------------------------------------------

BOOL parse_sepspec ()
{
	int col = 1;
	USHORT num;
	char *c = sepspec;
	unsigned short *sep;
	BOOL rval = TRUE, has_text;
	static unsigned short bit[] = {
		0x2000, 							// 0 row
		0x1000, 							// 1
		0x0800, 							// 2
		0x0400, 							// 3
		0x0200, 							// 4
		0x0100, 							// 5
		0x0080, 							// 6
		0x0040, 							// 7
		0x0020, 							// 8
		0x0010,								// 9 is bottom row
		0x0000,								// there is no 10 row
		0x4000,								// 11
		0x8000};							// 12 is top row

	ZeroMemory(sepbuf, sizeof(sepbuf));
	sep = (unsigned short *) sepbuf;						// alias as word array

	for (;;) {
		c = skipbl(c);										// skip leading blanks
		if (! *c)											// nothing left
			break;

		if (*c == ',') {									// assume commas mean a new column
			col++;
			c++;
			continue;
		}

		if (*c == '-')	{									// skip dashes
			c++;
			continue;
		}

		if (*c == '"') {									// it's a quoted string
			c++;

			while (*c && *c != '"') {
				num = ascii_to_hollerith(cur_encoding, *c++);

				if (col >= 1 && col <= 80) {
					sepbuf[col-1] = (USHORT) ((num == BAD_HOLLERITH) ? '?' : num);
					col++;
				}

				if (num == BAD_HOLLERITH) {
					rval = FALSE;
					break;
				}
			}
			if (! rval)
				break;

			if (*c == '"')									// skip over the closing quote
				c++;

			continue;
		}

		if (*c < '0' || *c > '9') {							// invalid character
			rval = FALSE;
			break;
		}

		num = 0;
		while (*c >= '0' && *c <= '9')
			num = num*10 + *c++ - '0';

		c = skipbl(c);
		if (*c == ':') {									// if followed by colon, it's a column number
			c++;
			if (num < 1 || num > 80) {
				rval = FALSE;
				break;
			}

			col = num;
		}
		else {												// else it's a row number
			if (num > 12 || num == 10) {
				rval = FALSE;
				break;
			}

			if (col >= 1 && col <= 80)
				sep[col-1] |= bit[num];						// set the specified bit
		}
	}

	if (autosep) {											// (condition on autosep so it won't force display mode
		CopyMemory(cardbuf, sepbuf, sizeof(cardbuf));		//  on when initializing the program, due to EN_CHANGE, if autosep is not enabled)
		convert_card();

		has_text = FALSE;
		for (c = cardtext; *c; c++) {
			if (*c != '?' && *c != ' ') {
				has_text = TRUE;
				break;
			}
		}

		if (! display)
			set_display_mode(TRUE, FALSE);

		show_card(TRUE, has_text);

		remark("Display shows autosep card");
	}

	return rval;
}

// -----------------------------------------------------------------
// increment_name - analyze current name and add 1 to it
// -----------------------------------------------------------------

void increment_name (void)
{
	char *c, fmt[10], nstr[10], fname[256];
	int n, ndig, nbase;

	if ((c = strrchr(curname, '\\')) == NULL)				// skip past path or device name
		if ((c = strrchr(curname, ':')) == NULL)
			c = curname;

	for (; *c && *c != '.' && ! isdigit(*c); c++)
		;

	if (! *c) {												// no previous number OR extension
		strcpy(fname, curname);
		strcat(fname, "1");
	}
	else {
		n = atoi(c)+1;										// new file number
		nbase = c-curname;									// number of characters to left of number
		strncpy(fname, curname, nbase);

		ndig = 0;											// number of digits in current name...
		while (isdigit(*c)) {
			c++;
			ndig++;
		}

		if (ndig == 0)										// none? use freeform
			sprintf(fmt, "%%d");
		else												// some? try to keep same number of digits (0 fill)
			sprintf(fmt, "%%0%dd", ndig);

		sprintf(nstr, fmt, n);								// format the new number and tack it onto base
		strcpy(fname+nbase, nstr);
		strcat(fname, c);									// append whatever came after numbers (e.g. extension)
	}

	SetDlgItemText(hDlgMain, IDC_FILENAME, fname);			// EN_CHANGE will ensue, setting curname
}

// -----------------------------------------------------------------
// browse_file - use standard SaveAs dialog to select output file
// -----------------------------------------------------------------

void browse_file (void)
{
	OPENFILENAME ofn;
	char namebuf[MAX_PATH];
	int i;
	DWORD attr;

	strcpy(namebuf, curname);
	if ((i = strlen(namebuf)) > 0)
		if (namebuf[i-1] == '\\')
			namebuf[i-1] = '\0';							// dialog doesn't like it if name ends in backslash

	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize       = sizeof(ofn);
	ofn.hwndOwner         = hDlgMain;
	ofn.hInstance         = ghInstance;
	ofn.lpstrFilter       = NULL;
	ofn.lpstrCustomFilter = NULL;
	ofn.nMaxCustFilter    = 0;
	ofn.nFilterIndex      = 0;
	ofn.lpstrFile         = namebuf;
	ofn.nMaxFile          = sizeof(namebuf);
	ofn.lpstrFileTitle    = NULL;
	ofn.lpstrInitialDir   = NULL;
	ofn.lpstrTitle        = NULL;
	ofn.Flags  		      = OFN_HIDEREADONLY|OFN_NOREADONLYRETURN|OFN_PATHMUSTEXIST;

	if (GetSaveFileName(&ofn)) {
		strcpy(curname, ofn.lpstrFile);

		attr = GetFileAttributes(curname);					// if it's a directory, add a \ to the end
		if (attr != 0xFFFFFFFF && (attr & FILE_ATTRIBUTE_DIRECTORY) != 0)
			strcat(curname, "\\");

		SetDlgItemText(hDlgMain, IDC_FILENAME, curname);	// this will cause EN_CHANGE which will call enable_controls()
	}
}

// -----------------------------------------------------------------
// read_file - start reading current file
// -----------------------------------------------------------------

void read_file (BOOL verify_mode)
{
	DWORD attr, threadid;
	char msg[MAX_PATH+100];

	did_exist = FALSE;										// verify before overwriting

	attr = GetFileAttributes(curname);
	did_exist = (attr != 0xFFFFFFFF);
	if (did_exist && (attr & FILE_ATTRIBUTE_DIRECTORY)) {
		MessageBox(hDlgMain, "Filename refers to a directory; please re-enter", NULL, MB_ICONSTOP|MB_OK|MB_APPLMODAL);
		return;
	}

	verify = verify_mode;

	if (verify) {
		if (! did_exist) {
			sprintf(msg, "File '%s' does not exist, can't verify", curname);
			MessageBox(hDlgMain, msg, NULL, MB_ICONSTOP|MB_OK|MB_APPLMODAL);
			return;
		}
		if ((fout = fopen(curname, to_ascii ? "r" : "rb")) == NULL) {
			remark("* Cannot access file");
			return;
		}
	}
	else {
		if (did_exist) {
			sprintf(msg, "File '%s' exists; do you want to overwrite it?", curname);
			if (MessageBox(hDlgMain, msg, "", MB_ICONSTOP|MB_YESNO|MB_APPLMODAL) != IDYES)
				return;
		}
	}

	if (hEvent == NULL) {									// create an event to synchronize the read thread
		if ((hEvent = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL) {
			bail("Unable to create signalling event");
			return;
		}
	}

	if (hThread == NULL) {									// create a separate thread to actually read the deck
		if ((hThread = CreateThread(NULL, 0, read_thread, NULL, 0, &threadid)) == NULL) {
			bail("Unable to create read thread");
			return;
		}
	}

	remark("%s \"%s\"", verify ? "Verifying" : "Reading", curname);	// emphasize name of file we're reading

	stop_pending  = FALSE;									// enter read mode:
	am_reading    = TRUE;									// set GUI to read mode
	do_autoincr   = do_autoverify = FALSE;

	show_card(FALSE, FALSE);								// blank the card display

	SetDlgItemText(hDlgMain, IDC_GO, "&Stop");				// turn Go button into Stop
	enable_controls();

	SetEvent(hEvent);										// unblock read thread, let it start running
}

// -----------------------------------------------------------------
// next_file - like read_file, but increments name first
// -----------------------------------------------------------------

void next_file (BOOL verify_mode)
{
	if (! *curname)
		strcpy(curname, "FILE000.DAT");

	if (am_reading) {
		do_autoincr   = TRUE;								// we are reading and hopper is empty;
		do_autoverify = FALSE;
		stop_pending  = TRUE;								// stop this file and start another
	}
	else {
		increment_name();
		read_file(verify_mode);
	}
}

// -----------------------------------------------------------------
// stop_reading - signal the read thread to stop as soon as possible
// -----------------------------------------------------------------

void stop_reading (BOOL from_EOF)
{
	do_autoverify = do_autoincr = FALSE;

	if (am_reading && from_EOF && nrec > 0)	{				// if a clean file ending occurred,
		if (autoverify && ! verify)
			do_autoverify = TRUE;
		else if (autoincr)									// see if we should automatically start a new read
			do_autoincr = TRUE;								// these do_xxx flags tell read thread to restart immediately
	}

	if (am_reading && ! stop_pending) {
		EnableControl(hDlgMain, IDC_GO, FALSE);				// disable the control to show we got the message
		stop_pending = TRUE;
	}
}

// -----------------------------------------------------------------
// read_data - read card data after seeing '='
// -----------------------------------------------------------------

BOOL read_data (void)
{
	DWORD nio;
	int i;

	assert(sizeof(cardbuf) == 160);

	ReadFile(hPort, cardbuf, 160, &nio, NULL);
	if (nio != 160) {
		remark("* Error reading data from controller");
		return FALSE;
	}

// Ver 1.2: convert card regardless of binary/ascii setting, if we're going to display it
	if (to_ascii || display)
		convert_card();

	if (autosep) {											// check to see if this is an autosep card
		for (i = 0; i < 80; i++)
			if (cardbuf[i] != sepbuf[i])
				break;

		is_autosep = (i >= 80);
	}
	else
		is_autosep = FALSE;

	if (display)
		show_card(TRUE, to_ascii);

	return TRUE;
}

// -----------------------------------------------------------------
// save_data - save current card to output file. It will have been
// converted to ascii, if necessary, by read_data
// -----------------------------------------------------------------

void save_data (void)
{
	BOOL failed;

	if (verify)
		return;

	if (fout == NULL) {
		if ((fout = fopen(curname, to_ascii ? "w" : "wb")) == NULL) {
			remark("* Can't open output file");
			stop_reading(FALSE);
			return;
		}
	}

	if (to_ascii) {
		failed = fputs(cardtext, fout) == EOF;
		if (! failed)
			failed = putc('\n', fout) == EOF;
	}
	else
		failed = fwrite(cardbuf, 2, 80, fout) != 80;

	if (failed) {
		stop_reading(FALSE);
		remark("* Error writing data to file");
	}
}

// -----------------------------------------------------------------
// view_file - open file with notepad
// -----------------------------------------------------------------

void view_file (void)
{
	char fname[MAX_PATH], cmd[2*MAX_PATH+10], *viewer;

	viewer = to_ascii ? ascii_viewer_path : binary_viewer_path;
	if (! *viewer)
		return;

	GetDlgItemText(hDlgMain, IDC_FILENAME, fname, sizeof(fname));
	fname[sizeof(fname)-1] = '\0';
	alltrim(fname);

	if (! *fname)											// no filename? ignore button (should not be enabled)
		return;

	sprintf(cmd, "\"%s\" \"%s\"", viewer, fname);
	WinExec(cmd, SW_SHOWNORMAL);
}

// -----------------------------------------------------------------
// ConfigurePortDlgProc - configure the IO port to the reader interface
// -----------------------------------------------------------------

BOOL CALLBACK ConfigurePortDlgProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			return HANDLE_WM_INITDIALOG(hDlg, wParam, lParam, ConfigurePort_OnInitDialog);

		case WM_COMMAND:
			HANDLE_WM_COMMAND(hDlg, wParam, lParam, ConfigurePort_OnCommand);
			break;

		default:
			return FALSE;
	}

	return TRUE;
}

// -----------------------------------------------------------------
// ConfigurePort_OnInitDialog - initialize program
// -----------------------------------------------------------------

BOOL ConfigurePort_OnInitDialog (HWND hDlg, HWND hwndFocus, LPARAM lParam)
{
	SetDlgItemText(hDlg, IDC_PORT, port);
	CheckDlgButton(hDlg, IDC_STOPONERROR, stoponerror);
	CenterMe(hDlg);											// center dialog on screen
	return TRUE;
}

// -----------------------------------------------------------------
// CardRead_OnCommand - handle command messages (GUI clicks)
// -----------------------------------------------------------------

void ConfigurePort_OnCommand (HWND hDlg, int id, HWND hWndCtl, UINT codeNotify)
{
	switch (id) {
		case IDOK:											// on OK, save port name
			GetDlgItemText(hDlg, IDC_PORT, port, sizeof(port));
			stoponerror = IsDlgButtonChecked(hDlg, IDC_STOPONERROR);
			EndDialog(hDlg, TRUE);
			break;

		case IDCANCEL:
			EndDialog(hDlg, FALSE);
			break;

		case IDC_PORT:
			if (codeNotify == EN_CHANGE) 					// disable OK if port name is blank
				EnableControl(hDlg, IDOK, SendDlgItemMessage(hDlg, IDC_PORT, WM_GETTEXTLENGTH, 0, 0) > 0);
			break;
	}
}

// -----------------------------------------------------------------
// set_display_mode - set new mode for card-display option 
// -----------------------------------------------------------------

void set_display_mode (BOOL newmode, BOOL do_paint)
{
	RECT rDlg, rCard;
	static int htDlg = 0;
	int wdDlg, newhtDlg;

	display = newmode;										// set display flag with new value

	GetWindowRect(GetDlgItem(hDlgMain, IDC_CARD), &rCard);
	GetWindowRect(hDlgMain, &rDlg);							// get screen coords of dialog, and of the card box inside it

	wdDlg = rDlg.right-rDlg.left;							// width of dialog

	if (htDlg == 0)											// on first call, save height of full dialog window (as created)
		htDlg = rDlg.bottom-rDlg.top;
															// set new window height
	newhtDlg = display ? htDlg : (rCard.top - rDlg.top + 6);

	SetWindowPos(hDlgMain, NULL, 0, 0, wdDlg, newhtDlg, SWP_NOMOVE|SWP_NOZORDER);

	if (display && do_paint && have_data) {					// if we're now displaying, update the text and binary display
		convert_card();
		show_card(TRUE, to_ascii);
	}
}

// -----------------------------------------------------------------
// show_card - update card image with current data
// -----------------------------------------------------------------

void show_card (BOOL show_data, BOOL show_text)
{
	have_data = show_data;									// if FALSE, card display is blanked
//	display_cardtext(show_text ? cardtext : "");
// ver 1.2: always display the card text
	display_cardtext(cardtext);
	InvalidateRect(hwCard, NULL, ! have_data);				// force repaint of the card image (no need to erase if data present, paint_card floods the rectangle)
}

// -----------------------------------------------------------------
// display_cardtext - set the card display's text to the specified string
// -----------------------------------------------------------------

void display_cardtext (char *txt)
{
	char textout[165], *tout;								// largest possible ouput string

	for (tout = textout; *txt; ) {							// protect &'s by doubling them
		if (*txt == '&')									// otherwise the dialog will display &x as an underlined x
			*tout++ = '&';

		*tout++ = *txt++;
	}
	*tout = '\0';

	SetDlgItemText(hDlgMain, IDC_CARDTEXT, textout);
}

// -----------------------------------------------------------------
// MyRectProc - superclass window procedure for dialog box "groupbox".
// This is used to intercept WM_PAINT message to the groupbox representing the card image
// -----------------------------------------------------------------

LRESULT CALLBACK MyRectProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LRESULT rval;

	rval = (*oldRectProc)(hWnd, uMsg, wParam, lParam);
	
	if (uMsg == WM_PAINT)
		paint_card();

	return rval;
}

// -----------------------------------------------------------------
// paint_card - paint the current cardbuf data in the IDC_CARD box; called
// after WM_PAINT is handled by the original control's window procedure.
// (This isn't a perfect situation, as it doesn't get to use the BeginPaint
// display context; it gets its own and so paints the entire region. But that
// isn't a terrible inefficiency as most paints will be the full region anyway).
// If verify is on, we paint red to show differences between the two decks:
// red punches where there was none in the file, and red outlines where
// there was a punch in the file but not in the current card
// -----------------------------------------------------------------

void paint_card (void)
{
	HDC hDC;
	RECT rCard, rBar, rHole;
	BOOL show_diffs;
	HPEN hOldPen = NULL;
	HBRUSH hOldBrush = NULL;
	int hCard, wCard, xCard, yCard, row, col, cdata, vdata, x, y[12], whole, hhole;
	enum {P_NONE, P_SOLIDBLACK, P_SOLIDRED, P_OPENRED} pmode;

	if (! display)											// The card image is hidden
		return;

	GetWindowRect(hwCard, &rCard);							// get screen coordinates for card and text boxes
	GetWindowRect(GetDlgItem(hDlgMain, IDC_BAR),  &rBar);

	hCard = rCard.bottom-rCard.top-(rBar.bottom-rBar.top)-1;// height of drawable space in card
	wCard = rCard.right-rCard.left-1;

	hDC = GetDC(hwCard);									// allocate a display context

	whole = (int) (wCard/80.*0.8) - 1;						// width and height of punch rectangles
	hhole = (int) (hCard/12.*0.7);

	xCard = 3;												// position of punch for column 1 row 12
	yCard = rBar.bottom - rCard.top + 2;

	rCard.left   = 1;
	rCard.right  = wCard - 1;
	rCard.bottom = rCard.bottom - rCard.top - 2;
	rCard.top    = yCard - 3;

	if (! have_data) {										// don't display data when a new read has been initiated
//		FillRect(hDC, &rCard, GetStockObject(LTGRAY_BRUSH));
		return;
	}

	show_diffs = verify && ! to_ascii && ! is_autosep;		// if TRUE, display differences in punches in red

	if (bManilla != NULL)									// paint the card region a nice manilla color
		FillRect(hDC, &rCard, bManilla);

	for (row = 0; row < 12; row++)							// precompute row y coordinates
		y[row] = yCard + (int) (row*hCard/12);

	if (show_diffs) {
		hOldPen   = SelectObject(hDC, pRed);				// select pen and brush in case we need to draw P_OPENRED
		hOldBrush = SelectObject(hDC, bManilla);
	}

	for (col = 0; col < 80; col++) {						// now, check every bit for a punch
		cdata = cardbuf[col];
		vdata = vbuf[col];
		x = xCard + (int) (wCard/80.*col);
		for (row = 0; row < 12; row++) {
			if (show_diffs) {								// in binary verify mode, display mismatches
				if (cdata & (0x8000 >> row))
					pmode = (vdata & (0x8000 >> row)) ? P_SOLIDBLACK : P_SOLIDRED;
				else
					pmode = (vdata & (0x8000 >> row)) ? P_OPENRED : P_NONE;
			}
			else
				pmode = (cdata & (0x8000 >> row)) ? P_SOLIDBLACK : P_NONE;

			if (pmode != P_NONE) {
				rHole.left   = x;
				rHole.right  = x + whole + 1;
				rHole.top    = y[row];
				rHole.bottom = y[row] + hhole + 1;

				switch (pmode) {
					case P_SOLIDBLACK:
						FillRect(hDC, &rHole, bBlack);
						break;
					case P_SOLIDRED:
						FillRect(hDC, &rHole, bRed);
						break;
					case P_OPENRED:
						Rectangle(hDC, rHole.left, rHole.top, rHole.right, rHole.bottom);
						break;
				}
			}
		}
	}

	if (hOldPen != NULL)									// restore original pen & brush before releasing DC
		SelectObject(hDC, hOldPen);
	if (hOldBrush != NULL)
		SelectObject(hDC, hOldBrush);

	ReleaseDC(hwCard, hDC);									// release the DC
}

// -----------------------------------------------------------------
// convert_card - convert binary data in cardbuf to text string in cardtext
// -----------------------------------------------------------------

void convert_card (void)
{
	int col;

	for (col = 0; col < 80; col++)							// get ascii equivalent for each character
		cardtext[col] = hollerith_to_ascii(cur_encoding, (USHORT) (cardbuf[col] & 0xFFF0));

	cardtext[col] = '\0';									// terminate string

	strim(cardtext);										// trim trailing blanks
}

// -----------------------------------------------------------------
// astring - allocate space and return a copy of a string
// -----------------------------------------------------------------

char *astring (char *str)
{
	char *s;

	if ((s = malloc(strlen(str)+1)) == NULL) {
		bail("Insufficient memory");
		return "";
	}

	strcpy(s, str);
	return s;
}

// -----------------------------------------------------------------
// alltrim - trim leading and trailing blanks from a string
// -----------------------------------------------------------------

char *alltrim (char *str)
{
	char *s;

	s = skipbl(str);

	if (s > str)											// if we skipped any, copy string down over the blank part
		strcpy(str, s);

	return strim(str);										// remove trailing spaces
}

// -----------------------------------------------------------------
// strim - remove trailing blanks from a string
// -----------------------------------------------------------------

char * strim (char *str)
{
	char *s, *sout;

	for (sout = str-1, s = str; *s; s++)
		if (*s > ' ')
			sout = s;

	sout[1] = '\0';
	return str;
}

// -----------------------------------------------------------------
// skipbl - skip leading whitespace in a string
// -----------------------------------------------------------------

char *skipbl (char *str)
{
	while (*str && *str <= ' ')
		str++;

	return str;
}

// -----------------------------------------------------------------
// issue_cmd - issue a command to the interface and get status response
// -----------------------------------------------------------------

BOOL issue_cmd (char cmd)
{
	DWORD nio, ioerr;
	char resp;
	int ntries = 10;
	BOOL resend = TRUE;

	while (--ntries >= 0) {
		if (resend && cmd != 0) {
			if (! debug_gui) {
				if (! WriteFile(hPort, &cmd, 1, &nio, NULL)) {
					remark("Error writing to interface");
					return FALSE;
				}
			}
		}

		resend = TRUE;

		if (debug_gui) {					// if we're debugging
			nio = 1;						// just fake a response
			resp = 'E';
		}
		else {
			nio = 0;
			ReadFile(hPort, &resp, 1, &nio, NULL);
		}

		if (nio != 1) {
			ClearCommError(hPort, &ioerr, NULL);
			continue;
		}

		if (resp == '?')
			continue;						// it didn't recognize the command

		if (valid_status(resp)) {
			status = resp;
			comm_ok = TRUE;					// enable status display
			if (am_reading)
				enable_controls();			// mmmm. status can change state of next button, need to do all of it
			update_status();
			return TRUE;
		}
		
		if (resp == '!') {					// we were expecting a response, but got a cancellation notice from pending pick
			resend = FALSE;					// status byte is coming
			ntries++;						// no penalty
			stop_reading(TRUE);				// notify reading routine that we are to stop
		}
		else if (resp == '=') {				// we were expecting a response, but are getting data from a pending pick
			read_data();
			resend = FALSE;					// status byte is coming
			ntries++;						// no penalty
			continue;
		}
		// else -- an unexpected character -- what to do? Ignore it? Guess so.
	}

	remark("Interface failed to respond after 10 tries");
	return FALSE;
}

// -----------------------------------------------------------------
// valid_status = return TRUE if status argument is a legal status byte
// -----------------------------------------------------------------

BOOL valid_status (char status)
{
	return status >= '@' && status <= '_';
}
// -----------------------------------------------------------------
// CenterMe - center calling window in its parent.  Call this from WM_INITDIALOG
// message for Dialog Boxes which we want centered on the screen.
// -----------------------------------------------------------------

void CenterMe (HWND hWnd)
{
	POINT pt;
	RECT  swp, rParent, rScreen;
	int   iwidth, iheight;
	HWND  hwParent, hwDesktop = GetDesktopWindow();

	if ((hwParent = GetParent(hWnd)) == NULL)
    	hwParent = hwDesktop;
	
	GetClientRect(hwParent, &rParent);						// get dimensions of parent window client area
	pt.x = rParent.right/ 2;								// find the center point in screen coordinates
	pt.y = rParent.bottom/ 2;
	ClientToScreen(hwParent, &pt);

	GetWindowRect(hWnd, &swp);								// get bounding rectangle of dialog
	iwidth  = swp.right - swp.left;							// calculate the height and width
	iheight = swp.bottom - swp.top;

	pt.x = pt.x - (iwidth / 2);								// put dialog center on center of parent client
	pt.y = pt.y - (iheight / 2);

	GetWindowRect(hwDesktop, &rScreen);						// get screen dimensions

	GetWindowRect(hwParent, &rParent);	

	if ((pt.x+iwidth) > rParent.right)						// if that would put part of dialog off
		pt.x = rParent.right-iwidth;						// the parent, move it back
	if (pt.x < rParent.left)
		pt.x = 0;

	if ((pt.y+iheight) > rParent.bottom)
		pt.y = rParent.bottom-iheight;
	if (pt.y < rParent.top)
		pt.y = 0;
															// move it to chosen position
	MoveWindow(hWnd, pt.x, pt.y, iwidth, iheight, FALSE);
}

// -----------------------------------------------------------------
// read_thread - thread that handles file reading (separate from GUI thread, so that
// relatively long times waiting for IO doesn't tie up the GUI).
//
// Sits here waiting for event to be signalled before proceeding.
// -----------------------------------------------------------------

DWORD CALLBACK read_thread (LPDWORD param)
{
	for (;;) {
		// wait for application to signal that we want to read a file
		WaitForSingleObject(hEvent, INFINITE);

		// read a file until told to quit
		read_one_file();									

		// restore normal GUI operation:
		issue_cmd('X');										// cancel pick
		issue_cmd('C');										// clear EOF lamp
		stop_pending = FALSE;
		am_reading   = FALSE;
		SetDlgItemText(hDlgMain, IDC_GO, "&Start");			// restore GUI to normal mode
		enable_controls();

		if (do_autoincr)									// start up again immediately?
			next_file(verify);								// (next_file or read_file will signal the event)
		else if (do_autoverify)
			read_file(TRUE);
	}														

	return 0;
}

// -----------------------------------------------------------------
// read_one_file = called from the read thread, reads one file until told to quit
// -----------------------------------------------------------------

void read_one_file (void)
{
	DWORD nio;
	char resp, msg[MAX_PATH+50], *nl, vstr[90];
	int i, ix, ch;
	BOOL picked = FALSE, trouble = FALSE, v_eof;

	issue_cmd('C');											// clear EOF lamp
	issue_cmd('S');											// get status
	SetDlgItemText(hDlgMain, IDC_COUNT, "");				// remove previous count & remark
	nrec = 0;												// clear card count

	for (;;) {												// wait for pick result character = or !
		if ((status & (STS_EOF | STS_EMPTY)) == (STS_EOF | STS_EMPTY))
			break;											// hopper is empty and EOF was pressed, so quit

		if (stop_pending)									// cancel request from GUI thread
			break;

		// I'm not issuing the first pick command until ready status comes up,
		// because the reader occasionally ignores the pick in this case, even
		// with the PC signal stuck on. I often had to hit stop and reset to get
		// the picking to start.
		//
		// BUT -- if we just got a successful pick, go ahead and 
		// issue the pick instantly, because, as it turns out, READY
		// does not come back on until about 3.5 msec after the last column was read,
		// and 200 msec is too long to wait!

		if ((status & STS_PENDING) == 0) {					// pick pending yet?
			if ((status & STS_READY) || picked)				// if ready, or if we just picked a card,
				issue_cmd('P');								// issue pick command
			else {
				Sleep(200);									// otherwise, wait a bit and check again
				issue_cmd('S');
			}
		}

		if ((status & STS_PENDING) == 0)					// if pick didn't stick, go try to issue it again
			continue;

		nio = 0;											// (timeout of 200 msec keeps this from spinning too fast)
		ReadFile(hPort, &resp, 1, &nio, NULL);				// wait for ! or = reply from pick command
		if (nio < 1)
			continue;										// no response yet, go back and check again

		if (resp == '!') {									// EOF button cancel
			issue_cmd(0);									// pick up status and stop
			stop_reading(TRUE);
			break;
		}
	
		if (resp == '=') {									// data follows
			SetDlgItemInt(hDlgMain, IDC_COUNT, ++nrec, FALSE);	// show card count

			if (verify && ! to_ascii)						// if we're verifying in binary mode, read the card record now
				v_eof = fread(vbuf, 2, 80, fout) != 80;		// so we can display differences (remember if we hit EOF)

			if (! read_data()) {							// read the 80 columns of data
				trouble = TRUE;								// oops, something went wrong (file io error maybe)
				break;
			}

			picked = TRUE;									// we've gotten our first card

			issue_cmd(0);									// pick up status

			if (onecard)
				break;

			if (is_autosep) {
				SetDlgItemInt(hDlgMain, IDC_COUNT, --nrec, FALSE);	// uncount it;

				if (nrec > 0 && autoincr) {					// if we saw cards, restart
					do_autoincr   = TRUE;
					do_autoverify = FALSE;					// we never autoverify with autosep
				}
				break;										// stop in any case (thus 2 autoseps together stop the reader)
			}

			if (verify) {									// data card, verify data
				if (to_ascii) {
					if (fgets(vstr, sizeof(vstr), fout) == NULL) {
						remark("* VERIFY FAILED: Too many cards");
						trouble = TRUE;						// EOF on file before end of cards
						break;
					}

					if ((nl = strchr(vstr, '\n')) != NULL)
						*nl = '\0';							// remove newline from text in file

					for (i = 0; vstr[i] != '\0'; i++)		// compare text
						if (cardtext[i] != vstr[i])
							break;

					if (vstr[i] != '\0') {
						vstr[i+4]     = '\0';				// show a few characters around the problem
						cardtext[i+4] = '\0';
						ix = (i > 3) ? (i-3) : i;
						remark("* VERIFY FAILED in col %d, file had \"%s\", read \"%s\"", i, cardtext+ix, vstr+ix);
						trouble = TRUE;
						break;
					}
				}
				else {										// binary verify: read data from file
					if (v_eof) {							// oops, EOF on file before end of cards?
						remark("* VERIFY FAILED - too many cards");
						trouble = TRUE;
						break;
					}

					for (i = 0; i < 80; i++)				// compare bytes
						if (cardbuf[i] != vbuf[i])
							break;

					if (i < 80) {							// oops, stopped before end
						remark("* VERIFY FAILED - Data differ in column %d", i+1);
						trouble = TRUE;
						break;
					}
				}
			}
			else											// save data
				save_data();

			if ((status & STS_ERROR) && stoponerror) {		// oops, we got a read check or something
				remark("* DATA ERROR -- POSSIBLY CORRUPT");
				trouble = TRUE;
				break;
			}

			if (verify && (status & STS_EMPTY)) {			// if hopper just emptied, see if file did too
				if ((ch = getc(fout)) == EOF)
					break;									// if last card was read AND file is at end of file, auto-quit

				ungetc(ch, fout);							// oops, file not empty. put the character back
			}
		}
		else {
			remark("* Unexpected pick command response: %c ($%0x)", resp, resp & 0xFF);
			trouble = TRUE;
			break;
		}
	}

	// end of read loop

	if (verify && nrec > 0 && ! trouble) {					// no trouble, so check for EOF on comparison file
		if (to_ascii ? (fgets(vstr, sizeof(vstr), fout) != NULL) : (fread(vbuf, 1, 160, fout) != 0)) {
			remark("* VERIFY FAILED - too few cards");
			trouble = TRUE;
		}
	}

	MessageBeep(trouble ? MB_ICONHAND : MB_OK);				// make happy or sad noise

	if (status & STS_PENDING)
		issue_cmd('X');										// cancel pending pick (if we broke out due to error)

	if (fout != NULL) {										// close file, if open
		fclose(fout);
		fout = NULL;
	}
															// if remark is still blank or the initial message
															// offer something useful
	GetDlgItemText(hDlgMain, IDC_REMARK, msg, sizeof(msg));
	if (strncmp(msg, "Reading \"", 9) == 0 || strncmp(msg, "Verifying \"", 9) == 0 || ! *msg) {
		if (verify)
			remark((nrec == 0) ? "(Verify canceled)" : "* verify OK");
		else if (nrec == 0) {
			if (did_exist)
				remark("(Read canceled, original file left alone)");
			else
				remark("(Read canceled, file not created)");
		}
		else {
			remark("%d card%s read", nrec, (nrec == 1) ? "" : "s");
		}

		SetDlgItemText(hDlgMain, IDC_COUNT, "");			// remove count display
	}

	if (trouble || nrec == 0)
		do_autoincr = do_autoverify = FALSE;
}
